import * as entityForm from './edit_plus/entity-form.js';
import * as editableElement from './edit_plus/editable-element.js';
import * as pluginManager from './edit_plus/field-plugin-manager.js';

(($, Drupal, once) => {

  Drupal.EditPlus = {};
  Drupal.EditPlus.EditMode = 'enabled';

  /**
   * Focused Form Item.
   *
   * Remember the focused form item. Instead of calling replaceEditableElementWithMarkup
   * onblur, lets track the focused form item so that clicking things in the
   * sidebar doesn't lose focus on the currently edited form item.
   */
  Drupal.EditPlus.Focused = null;

  /**
   * Only update element.
   *
   * Flag to only update the markup of the changed field. Usually after an
   * element has been edited the entire entity's rendered markup is updated. If
   * a user edit's an element and then clicks another editable element which is
   * then replaced with the form item for editing, the form item would be
   * overwritten/disappear when the whole entity's markup is replaced. In this
   * instance only update the field that has changed.
   *
   * @type {null}
   */
  Drupal.EditPlus.onlyUpdateElement = null;

  /**
   * Editable Element clicked.
   *
   * @param e
   *   The click event.
   */
  Drupal.EditPlus.EditableElementClicked = (e) => {
    // Fields are often nested. Only register one mousedown.
    e.preventDefault();
    e.stopPropagation();
    // Deselect any currently focused fields.
    if (Drupal.EditPlus.Focused) {
      Drupal.EditPlus.FieldPluginManager.getPlugin(Drupal.EditPlus.Focused.elementInfo).blur(Drupal.EditPlus.Focused);
    }

    // @todo Make the default of which form is loaded initially configurable and
    //  add a UI to switch between the entity reference forms.
    let fieldValue = e.target.closest('.edit-plus-field-value');
    while (fieldValue.parentNode.closest('.edit-plus-field-value')) {
      fieldValue = fieldValue.parentNode.closest('.edit-plus-field-value');
    }

    const EditableElement = new editableElement.EditableElement(fieldValue.dataset.editPlusId);

    entityForm.getForm(EditableElement).then((response, status) => {
      EditableElement.setWidget();
      Drupal.EditPlus.FieldPluginManager.getPlugin(EditableElement).edit(EditableElement);

      // if (response.status === 'new_form') {
      //   entityForm.autosaveNonEditableElements(EditableElement);
      // }
    });
  }

  /**
   * Enable edit mode.
   */
  Drupal.EditPlus.EnableEditMode = () => {
    once.remove('EditPlusDisabled', '.edit-plus-field-value', document);
    once('EditPlusEnabled', '.edit-plus-field-value', document).forEach(editableField => {
      editableField.addEventListener('mousedown', Drupal.EditPlus.EditableElementClicked);
      // Temporarily reveal the sidebar here. // @todo move to toolbar+
      const sidebar = document.querySelector('#toolbar-plus-sidebar');
      if (sidebar.querySelector('.edit-plus-form')) {
        sidebar.style.display = 'inherit';
      }
    });
    Drupal.EditPlus.DisableInteractiveElements();
  }

  /**
   * Disable edit mode
   */
  Drupal.EditPlus.DisableEditMode = () => {
    once.remove('EditPlusEnabled', '.edit-plus-field-value', document);
    once('EditPlusDisabled', '.edit-plus-field-value', document).forEach(editableField => {
      editableField.removeEventListener('mousedown', Drupal.EditPlus.EditableElementClicked);
    });
    document.querySelector('#toolbar-plus-sidebar').style.display = 'none'; // @todo move to toolbar+
    Drupal.EditPlus.EnableInteractiveElements();
  }

  /**
   * Toggle edit mode.
   *
   * This listens for toolbar_plus's toggleToolbarTool event.
   */
  Drupal.behaviors.EditPlusToggle = {
    attach: (context, settings) => {
      once('EditPlusToolbarToggle', 'html', context).forEach(element => {
        // Toggle Edit+ on and off.
        window.addEventListener('toggleToolbarTool', e => {
          if (e.detail.tool === 'edit_plus' && e.detail.state === 'on') {
            Drupal.EditPlus.EnableEditMode();
          } else {
            Drupal.EditPlus.DisableEditMode();
          }
        });
      });
    }
  };

  /**
   * Edit + persist state.
   *
   * Toolbar + dispatches events for when Edit mode should be enabled, but we
   * need to persist the state through AJAX calls.
   */
  Drupal.behaviors.EditPlusPersistState = {
    attach: (context, settings) => {
      once('EditPlusPersistState', '.edit-plus-entity-wrapper', context).forEach(element => {
        // Persist the state.
        if (
          sessionStorage.getItem('toolbarState') === 'open' &&
          localStorage.getItem('toolbar_plus.tool.active') === 'edit_plus'
        ) {
          Drupal.EditPlus.EnableEditMode();
        }
      });
    }
  };


  pluginManager.default($, Drupal, once);

  /**
   * Disables interactive elements while in edit mode.
   */
  Drupal.EditPlus.DoNothing = (e) => {
    e.preventDefault();
    e.stopPropagation();
  };
  Drupal.EditPlus.DisableInteractiveElements = () => {
    $('.edit-plus-entity-wrapper')
      .find('a')
      .each((index, link) => {
        // Disable clicking links while in edit mode.
        $(link).on('click touchstart', Drupal.EditPlus.DoNothing);
      });
  };

  Drupal.EditPlus.EnableInteractiveElements = () => {
    $('.edit-plus-entity-wrapper')
      .find('a')
      .each((index, link) => {
        // Enable clicking links when leaving edit mode.
        $(link).off('click touchstart', Drupal.EditPlus.DoNothing);
      });
  };

})(jQuery, Drupal, once);




