import * as util from './utilities.js';

/**
 * Editable Element.
 *
 * This class is a state object that represents the binding of the currently
 * edited page element to the form element.
 */
export class EditableElement {

  constructor(id) {
    this.info = util.parseMarkupId(id)
  }

  /**
   * Get form.
   *
   * @returns {Element}
   * The form for the EditableElement.
   */
  getForm() {
    return document.querySelector('[data-edit-plus-form-id="' + util.getFormId(this.info) + '"]');
  }

  setWidget() {
    this.info.widget = this.getFormItemWrapper().dataset.editPlusFormItemWidget;
  }

  /**
   * Get page element.
   *
   * @returns {Element}
   *   The rendered markup on the page for the EditableElement.
   */
  getPageElement() {
    return document.querySelector('[data-edit-plus-id="' + this.info.elementId + '"]');
  }

  getPageElementWrapper() {
    return document.querySelector('[data-edit-plus-field-value-wrapper="' + util.getPageElementWrapperId(this.info) + '"]');
  }

  getPageElementHandle() {
    if (this.getHandleType() === 'wrapper') {
      return this.getPageElementWrapper();
    }
    return document.querySelector('[data-edit-plus-id="' + this.info.elementId + '"]');
  }

  /**
   * Get form item.
   *
   * @returns {Element}
   * The form item for the EditableElement.
   */
  getFormItem() {
    return document.querySelector('[data-edit-plus-form-item-id="' + util.getFormItemId(this.info) + '"]');
  }

  /**
   * Get form item wrapper.
   *
   * @returns {Element}
   *   The form item wrapper for the EditableElement.
   */
  getFormItemWrapper() {
    return document.querySelector('[data-edit-plus-form-item-wrapper-id="' + util.getFormItemWrapperId(this.info) + '"]');
  }

  /**
   * Get form item handle.
   *
   * @returns {Element}
   *   The handle for the form element that will be swapped in for inline editing.
   *   @see getHandleType.
   */
  getFormItemHandle() {
    if (this.getHandleType() === 'wrapper') {
      return this.getFormItemWrapper();
    }
    return document.querySelector('[data-edit-plus-form-item-id="' + util.getFormItemId(this.info) + '"]').closest('.edit-plus-form-item');
  }

  /**
   * Get handle type.
   *
   * @returns {string}
   *   Returns a string of either form_item or wrapper.
   *
   *   Users can config whether a field swaps the form item or the entire form
   *   wrapper into the page for editing.
   */
  getHandleType() {
    return this.getFormItemWrapper().dataset.editPlusHandle ?? 'form_item';
  }

  /**
   * Get form input.
   *
   * @returns {Element}
   *   The input element for the EditableElement.
   */
  getFormItemInputs() {
    return this.getFormItemHandle().querySelectorAll('[data-edit-plus-input]');
  }

  /**
   * Get form place marker.
   *
   * @returns {Element}
   *   The place marker element for the EditableElement.
   *
   *   When a form item is moved from the form to the page for editing a
   *   placeholder is created to mark where the form item originated from.
   */
  getFormPlaceMarker() {
    return document.querySelector('[data-form-item-placeholder="' + util.getFormItemWrapperId(this.info) + '"]');
  }

  /**
   * Is cloneable.
   *
   * @returns {boolean}
   *   Whether this field has been configured to swap a clone or the actual
   *   form item into the page for editing.
   */
  isCloneable() {
    return this.getFormItemWrapper().dataset.editPlusClone === 'true';
  }

  /**
   * Handle is wrapper.
   *
   * @see getFormItemHandle
   *
   * @returns {boolean}
   *   Returns true if the handle is wrapper. false if the handle is form_item.
   */
  handleIsWrapper() {
    const wrapper = this.getFormItemWrapper();
    const elementToUseAsHandle = wrapper.dataset.editPlusHandle ?? 'form_item';
    return elementToUseAsHandle === 'wrapper';
  }

  /**
   * Create from form item.
   *
   * @param element
   *   An element within a form item wrapper.
   *
   * @returns {EditableElement}
   *   A new EditableElement instance from a form item or any element from
   *   within a form item wrapper.
   */
  static createFromFormItem(element) {
    const wrapper = element.closest('.edit-plus-form-item-wrapper');
    if (wrapper) {
      return new EditableElement(wrapper.dataset.editPlusPageElementId);
    }
    const pageElement = document.querySelector('[data-edit-plus-markup-item-id="' + element.dataset.editPlusFormItemId + '"]');
    return new EditableElement(pageElement.dataset.editPlusId);
  }

}
