import * as utilities from './utilities.js';

/**
 * Get form.
 *
 * @param EditableElement
 *   The element info object.
 * @returns {Promise<unknown>}
 *   A promise that retrieves either an Entity Form, or a Layout Builder (block)
 *   entity update form. If the form has already been fetched then it reveals
 *   the appropriate form.
 */
export const getForm = (EditableElement) => {
  // Hide/cache previously loaded forms.
  document.querySelectorAll('.edit-plus-form').forEach(form => {
    form.classList.add('edit-plus-hidden');
  });
  // Return the requested form if we already have it.
  const formExists = EditableElement.getForm();
  if (formExists) {
    return new Promise((resolve, reject) => {
      formExists.classList.remove('edit-plus-hidden');
      resolve({status: 'existing_form'});
    });
  }

  // Fetch a new form.
  const pageElement = EditableElement.getPageElement();
  const layoutBuilderBlock = pageElement.closest('.block-layout-builder');
  if (layoutBuilderBlock) {
    // Use the regular entity form for field blocks.
    // @todo Maybe add a field-block class that we don't need to search for...
    let isFieldBlock = false;
    for (let i=0; i < layoutBuilderBlock.classList.length; i++) {
      isFieldBlock = layoutBuilderBlock.classList[i].includes('field-block');
      if (isFieldBlock) {
        break;
      }
    }
    if (!isFieldBlock) {
      return getLayoutBuilderEntityForm(EditableElement);
    }
  }
  return getEntityForm(EditableElement);
}

/**
 * Get Layout Builder Entity Form.
 *
 * @param EditableElement
 *   The element info object.
 *
 * @returns {Promise<unknown>}
 *   A promise to retrieve the form.
 */
const getLayoutBuilderEntityForm = (EditableElement) => {
  return new Promise((resolve, reject) => {
    // Figure out the URL parameters based on the markup.
    const pageElement = EditableElement.getPageElement();
    const storageIds = jQuery(pageElement).parents('[data-layout-builder-section-storage-id]');
    const sectionStorage = storageIds[storageIds.length - 1].dataset
    const storageType = sectionStorage.layoutBuilderSectionStorageType;
    const storageId = sectionStorage.layoutBuilderSectionStorageId;
    const region = pageElement.closest('[data-layout-builder-region]').dataset.layoutBuilderRegion;
    const block = pageElement.closest('[data-layout-builder-block-uuid]');
    const sectionDelta = block.closest('[data-layout-builder-section-delta]').dataset.layoutBuilderSectionDelta;
    const blockUuid = block.dataset.layoutBuilderBlockUuid;
    let nestedStoragePath = '';
    for (let i = storageIds.length - 2; i >= 0; i--) {
      const layoutBlock = storageIds[i].closest('[data-layout-builder-layout-block]');
      const section = layoutBlock.closest('[data-layout-builder-section-delta]');
      nestedStoragePath = `${section.dataset.layoutBuilderSectionDelta}&${layoutBlock.dataset.layoutBuilderBlockUuid}` + nestedStoragePath;
    }
    nestedStoragePath = nestedStoragePath ? '/' + nestedStoragePath : nestedStoragePath;
    let url = Drupal.url(`edit-plus/update/block/${storageType}/${storageId}/${sectionDelta}/${region}/${blockUuid}${nestedStoragePath}`);

    EditableElement.info.ajaxReturnForm = true;
    let ajax = Drupal.ajax({
      url: url,
      type: 'GET',
      dataType: 'text',
      submit: EditableElement.info,
      progress: {
        type: 'fullscreen',
        message: Drupal.t('Loading form...')
      },
      error: error => {
        console.error(error.responseText);
        reject(error);
      },
      success: (response, status) => {
        Promise.resolve(
          response.forEach(r => {
            // @todo Is setting the selector here a hack? Consider making a renderer like Drupal\Core\Render\MainContent\AjaxRenderer.
            if (r.command === 'insert' && r.method === null && r.data && r.data.includes('edit-plus-form')) {
              r.method = 'append';
              r.selector = '#toolbar-plus-sidebar';
            }
          }),
        )
          .then(() => {
            Drupal.Ajax.prototype.success.call(ajax, response, status).then(() => {
              resolve({status: 'new_form'});
            });
          })
          .then(() => {
            jQuery('#toolbar-plus-sidebar').show();
          });
      }
    });
    ajax.execute();
    EditableElement.info.ajaxReturnForm = false;
  });
}

/**
 * Get Entity Form.
 *
 * @param EditableElement
 *   The element info object.
 *
 * @returns {Promise<unknown>}
 *   A promise to retrieve the form.
 */
const getEntityForm = (EditableElement) => {
  return new Promise((resolve, reject) => {

    EditableElement.info.ajaxReturnForm = true;
    let ajax = Drupal.ajax({
      url: Drupal.url('edit-plus/entity-form/' + EditableElement.info.entityTypeId + '/' + EditableElement.info.entityId),
      type: 'POST',
      dataType: 'text',
      submit: EditableElement.info,
      progress: {
        type: 'fullscreen',
        message: Drupal.t('Loading form...')
      },
      error: error => {
        console.error(error.responseText);
        reject(error);
      },
      success: (response, status) => {
        Promise.resolve(
          Drupal.Ajax.prototype.success.call(ajax, response, status)
        ).then(() => {
          resolve({status: 'new_form'});
        });
      }
    });
    ajax.execute();
    EditableElement.info.ajaxReturnForm = false;
  });
}

// @todo I think we need to tag auto save items in the sidebar on the backend.
// export const autosaveNonEditableElements = (EditableElement) => {
//   EditableElement.getForm().querySelectorAll('[data-edit-plus-input]').forEach(input => {
//     const formItem = input.closest('[data-edit-plus-form-item-id]');
//     if (formItem) {
//       const pageElement = utilities.getMarkupItem(formItem);
//       if (!pageElement) {
//         // Auto save changes to sidebar form items.
//         input.addEventListener('change', e => {
//           const updateButton = input.closest('form').querySelector('.edit-plus-update-button');
//           jQuery(updateButton).mousedown();
//           editPlusIsUpdating = true;
//         });
//       }
//     }
//   });
// }

export let editPlusIsUpdating = false;

/**
 * Update tempstore.
 *
 * @param EditableElement
 *   The form item whose form needs updating.
 * @param isEmpty
 *   Whether the field has been emptied or not. Emptied fields are moved to the
 *   sidebar.
 */
export const updateTempstore = (EditableElement, isEmpty) => {
  const form = EditableElement.getFormItemWrapper().closest('form');

  // Move emptied fields to the sidebar.
  if (isEmpty) {
    form.querySelector('#empty-field').value = EditableElement.info.fieldName;
  }

  // Click the update button (formerly the save submit).
  const updateButton = form.querySelector('.edit-plus-update-button');
  jQuery(updateButton).mousedown();
  editPlusIsUpdating = true;
}

