import * as editableElement from '../editable-element.js';

/**
 * Field plugin base.
 *
 * A base class for field plugins.
 */
export class FieldPluginBase {

  constructor() {
    this.isObserverActive = false;
    this.originalValue = null;
    this.clonedElement = null;
  }

  init() {}

  /**
   * Supported widgets.
   *
   * @type {[]}
   *   An array of field formatter ID's this plugin supports.
   */
  supportedWidgets = [];

  /**
   * Edit.
   *
   * @param EditableElement
   *   The element being edited.
   */
  edit(EditableElement) {
    this.replacePageMarkupWithEditableElement(EditableElement);
    this.focus(EditableElement);
    this.revealAncillary(EditableElement);
  }

  /**
   * Focus.
   *
   * Focus on the field after it has been clicked.
   *
   * @param EditableElement
   *   The element being edited.
   */
  focus(EditableElement) {
    // Prepare for de-focusing the form item. It would be easy to add a
    // formItem.onblur when the form item looses focus, but losing focus on the
    // form item doesn't necessarily mean we are done with it. e.g. changing a
    // text format in the sidebar.
    EditableElement.getFormItemWrapper().classList.add('edit-plus-focused');
    Drupal.EditPlus.Focused = {
      formItem: EditableElement.getFormItemHandle(),
      elementInfo: EditableElement.info,
      // @todo move the above to instead use Editable Element.
      editableElement: EditableElement,
    };
    document.addEventListener('mousedown', this.watchFocus, true);

    this.focusElement(EditableElement)
  }

  /**
   * Focus element.
   *
   * This should be overridden by field plugins to provide field specific
   * instructions on how to give the element focus.
   *
   * @param EditableElement
   *
   * @param event
   */
  focusElement(EditableElement, event) {}

  /**
   * Watch focus.
   *
   * Once an element is being edited, monitor clicks to determine if the click
   * should cause the form item to lose focus. Field plugins can provide selectors
   * that do not make the form item loose focus when clicked. e.g. things in the
   * sidebar like field formatters etc.
   *
   * @param e
   *   The mousedown event.
   */
  watchFocus(e) {
    // Check that the clicked element is something that should cause the form
    // item to lose focus.
    const plugin = Drupal.EditPlus.FieldPluginManager.getPlugin(Drupal.EditPlus.Focused.editableElement);
    const getCurrentlyFocusedSelectors = plugin.getFocusedSelectors();
    let remainFocused = false;
    for (let i = 0; i < getCurrentlyFocusedSelectors.length; i++) {
      const ignoredClick = e.target.closest(getCurrentlyFocusedSelectors[i]);
      if (ignoredClick !== null) {
        remainFocused = true;
        break;
      }
    }
    if (!remainFocused) {
      const fieldValue = e.target.closest('.edit-plus-field-value');
      if (fieldValue) {
        // We lost focus when clicking another editable element. Only update the
        // rendered page with the last changed element.
        Drupal.EditPlus.onlyUpdateElement = Drupal.EditPlus.Focused.editableElement;
      }

      plugin.blur();
      document.querySelectorAll('.edit-plus-focused').forEach(element => {
        element.classList.remove('edit-plus-focused');
      });
    }
  }

  /**
   * Blur.
   *
   * Remove focus from the actively focused form item.
   */
  blur() {
    const EditableElement = Drupal.EditPlus.Focused.editableElement;
    if (Drupal.EditPlus.Focused.editableElement.isCloneable()) {
      this.isObserverActive = false;
      this.observer.disconnect();
    }
    document.removeEventListener('mousedown', this.watchFocus, true);
    Drupal.EditPlus.Focused = null;
    this.blurElement(EditableElement);
  }

  /**
   * Blur element.
   *
   * This should be overridden by field plugins to provide field specific
   * instructions on how to remove focus from the element. Since detecting if
   * the field has changes after editing it's recommended for field plugins to
   * call this.replaceEditableElementWithMarkup(EditableElement) when appropriate.
   *
   * @param EditableElement
   *   The element being edited.
   */
  blurElement(EditableElement) {}

  /**
   * Reveal ancillary.
   *
   * Reveal things in the sidebar that pertain to this field e.g. formatter
   *
   * @param EditableElement
   *   The element being edited.
   */
  revealAncillary(EditableElement) {}

  /**
   * Hide ancillary.
   *
   * Reveal things in the sidebar that pertain to this field e.g. formatter
   *
   * @param EditableElement
   *   The current form item.
   */
  hideAncillary(EditableElement) {
    EditableElement.getFormItemWrapper().classList.add('edit-plus-hidden');
  }


  /**
   * Replace page markup with editable element.
   *
   * Moves a form item from the form to the main page area. Users can configure
   * whether the actual form item is swapped for the rendered element on the page
   * or a clone of the form item. Some form items like media libraries stop
   * working when taken out of the form. In that case a clone is created and the
   * clicks on the clone are forwarded to the original form item.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replacePageMarkupWithEditableElement(EditableElement) {
    if (EditableElement.isCloneable()) {
      this.replacePageMarkupWithClonedFormItem(EditableElement);
    } else {
      this.replacePageMarkupWithFormItem(EditableElement);
    }
  }

  /**
   * Replace page markup with form item.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replacePageMarkupWithFormItem(EditableElement) {
    const pageElementHandle = EditableElement.getPageElementHandle();
    const formItemWrapper = EditableElement.getFormItemWrapper();
    const formElementHandle = EditableElement.getFormItemHandle();
    const placeMarker = document.createElement('div');
    placeMarker.dataset.formItemPlaceholder = formItemWrapper.dataset.editPlusFormItemWrapperId;
    formElementHandle.parentNode.insertBefore(placeMarker, formElementHandle);
    pageElementHandle.classList.add('edit-plus-hidden');
    pageElementHandle.parentNode.insertBefore(formElementHandle, pageElementHandle)
    formElementHandle.appendChild(pageElementHandle);
    formElementHandle.classList.remove('edit-plus-hidden');
    formElementHandle.classList.add('edit-plus-focused');
  }

  /**
   * Replace page markup with cloned form item.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replacePageMarkupWithClonedFormItem(EditableElement) {
    this.isObserverActive = true;

    const pageElementHandle = EditableElement.getPageElementHandle();
    pageElementHandle.classList.add('edit-plus-hidden');

    const clone = this.cloneElement(EditableElement);
    pageElementHandle.parentNode.insertBefore(clone, pageElementHandle);
    clone.classList.remove('edit-plus-hidden');
  }

  /**
   * Clone element.
   *
   * @param EditableElement
   *   The element being edited.
   *
   * @returns {Node}
   *   A clone of the EditableElement form item's handle.
   */
  cloneElement(EditableElement) {
    const formElementHandle = EditableElement.getFormItemHandle();
    // Remove old clones.
    document.querySelectorAll('#' + this.getClonedElementId(formElementHandle)).forEach(clone => {
      clone.remove();
    });
    // Make a clone of the form item formElementHandle which will be placed in the page for
    // editing.
    const clone = formElementHandle.cloneNode(true);
    // Make an additional clone to remember the current state of the element to
    // later detect if any AJAX changes have occurred to the original form item
    // so that we can only update the page when necessary.
    this.clonedElement = formElementHandle.cloneNode(true);

    // Prepare the cloned form element for being placed on the page by tweaking
    // its attributes so the clone is no longer a duplicate of the form item
    // which might confuse JS.
    clone.id = this.getClonedElementId(clone);
    clone.setAttribute('data-edit-plus-clone-id', 'handle');
    const handleChildren = formElementHandle.querySelectorAll('*');
    const cloneChildren = clone.querySelectorAll('*');
    handleChildren.forEach((element, index) => {
      // Identify it as a clone.
      if (element.id) {
        cloneChildren[index].id = this.getClonedElementId(element);
      }

      // Ensure we have unique ID's for every div that correlates it back to the
      // original so we can transfer clicks from the clone back to the original.
      if (element.hasAttribute('data-edit-plus-original-id')) {
        element.removeAttribute('data-edit-plus-original-id');
      }
      element.setAttribute('data-edit-plus-original-id', index);
      cloneChildren[index].setAttribute('data-edit-plus-clone-id', index);

      // Remove any cloned fadeIn's.
      if (element.style.opacity) {
        element.style.opacity = '';
      }
    });

    // Transfer events on the cloned element (clicks, drags, input, etc) to the
    // original form element.
    Object.keys(window).forEach(key => {
      if (/^on(?!animationiteration|animationstart|mouseenter|mouseout|mouseevent|mousemove|mouseover|mouseleave|mousein|pointerenter|pointerleave|pointerout|pointerup|pointermove|pointerrawupdate|pointerover|pointerdown|transitionrun|transitionstart|transitioncancel|transitionend|scroll|scrollend)/.test(key)) {
        clone.addEventListener(key.slice(2), e => {
          const cloneId = e.target?.dataset?.editPlusCloneId ?? false;
          if (!cloneId) {
            return;
          }
          const originalFormElement = this.getOriginalFromClone(cloneId);
          if (e.type === 'input') {
            // Transfer input changes.
            const clonedInput = e.target;
            const originalInput = this.getOriginalFromClone(clonedInput.dataset.editPlusCloneId);
            // If the event is a select list change. Update the selected
            // attribute so that when it is cloned again the selected option persists.
            if (clonedInput.nodeName === 'SELECT') {
              originalInput[originalInput.selectedIndex].removeAttribute('selected');
              originalInput[clonedInput.selectedIndex].setAttribute('selected', 'selected');
            }
            originalInput.value = clonedInput.value;
          } else {
            // Transfer any other events.
            const forwardEvent = jQuery.Event(e.type);
            forwardEvent.originalEvent = e;
            jQuery(originalFormElement).trigger(forwardEvent);
          }
        });
      }
    });

    clone.dataset.editPlusClone = formElementHandle.dataset.editPlusFormItemWrapperId;

    // Pass along the Edit+ ID so recreate EditableElement as changes occur
    formElementHandle.dataset.editPlusPageElementId = EditableElement.getPageElement().dataset.editPlusId;

    // Watch for changes to the original element.
    this.observer = new MutationObserver((mutations) => {
      if (!this.isObserverActive) {
        return;
      }
      const EditableElement = editableElement.EditableElement.createFromFormItem(mutations[0].target);

      // Update the clone if there has been changes to the form element.
      if (this.clonedElement.outerHTML !== EditableElement.getFormItemHandle().outerHTML) {
        this.replacePageMarkupWithEditableElement(EditableElement);
      }
    });
    this.observer.observe(formElementHandle, {
      attributes: true,
      attributeFilter: ['checked'],
      childList: true,
      subtree: true,
    });

    // Mutation observers can't detect checkboxes. Watch for those changes explicitly.
    formElementHandle.querySelectorAll('input[type="checkbox"]').forEach(checkbox => {
      checkbox.addEventListener('change', e => {
        const EditableElement = editableElement.EditableElement.createFromFormItem(e.target);
        this.replacePageMarkupWithEditableElement(EditableElement);
      });
    });

    return clone;
  }


  getOriginalFromClone(cloneId) {
    return document.querySelector('[data-edit-plus-original-id="' + cloneId + '"]');
  }

  /**
   * Get cloned element ID.
   *
   * @param element
   *  A div with an ID inside the form item handle being cloned.
   *
   * @returns {string}
   *   An edit plus cloned + id string.
   */
  getClonedElementId(element) {
    return 'edit-plus-cloned_' + element.id;
  }

  /**
   * Replace form item with markup.
   *
   * Moves a form item back to the form from the main page area.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replaceEditableElementWithMarkup(EditableElement) {
    if (EditableElement.isCloneable()) {
      this.replaceClonedFormItemWithPageElement(EditableElement);
    } else {
      this.replaceFormItemWithPageElement(EditableElement);
    }
  }

  /**
   * Replace cloned form item with page element.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replaceClonedFormItemWithPageElement(EditableElement) {
    const formElementHandle = EditableElement.getFormItemHandle();

    // Remove old clones.
    document.querySelectorAll('#' + this.getClonedElementId(formElementHandle)).forEach(clone => {
      clone.remove();
    });
    // Remove binding attribute.
    document.querySelectorAll('[data-edit-plus-original-id]').forEach(element => {
      element.removeAttribute('data-edit-plus-original-id');
    })
    EditableElement.getPageElementHandle().classList.remove('edit-plus-hidden')
  }

  /**
   *  Replace form item with page element.
   *
   * @param EditableElement
   *   The element being edited.
   */
  replaceFormItemWithPageElement(EditableElement) {
    const pageElementHandle = EditableElement.getPageElementHandle();
    const placeMarker = EditableElement.getFormPlaceMarker();
    const formElementHandle = EditableElement.getFormItemHandle();
    formElementHandle.parentNode.insertBefore(pageElementHandle, formElementHandle);
    placeMarker.parentNode.insertBefore(formElementHandle, placeMarker);
    pageElementHandle.classList.remove('edit-plus-hidden')
    placeMarker.remove();
    Drupal.EditPlus.FieldPluginManager.getPlugin(EditableElement).hideAncillary(EditableElement);
  }

  /**
   * Get focused selectors.
   *
   * When an input like a text area or an inline editor has focus, we don't always
   * want to hide the editor when the user clicks on something else on the page.
   * e.g. clicking a Bold icon in the CKEditor or changing a text format in the
   * sidebar.
   *
   * @returns {string[]}
   *   An array of selectors that will not cause the current form item to loose
   *   focus if clicked.
   */
  getFocusedSelectors() {
    return [
      '.edit-plus-focused',
      '[data-edit-plus-clone]',
    ];
  }

}

