import * as util from '../utilities.js';
import * as entityForm from '../entity-form.js';
import * as fieldPluginBase from './field-plugin-base.js';
import * as editableElement from '../editable-element.js';
import * as core from '../copied_from_core/copied-from-core.js';


export default function ($ = jQuery, Drupal, once, dropZones) {

  /**
   * InlineEditor field plugin.
   */
  class InlineEditor extends fieldPluginBase.FieldPluginBase {

    /**
     * Supported widgets.
     *
     * @type {[string]}
     */
    supportedWidgets = [
      'inline_textarea',
    ];

    /**
     * Original data.
     *
     * The original input so we can determine if it's been changed.
     *
     * @type {Map<any, any>}
     */
    originalData = new Map();

    focusElement(EditableElement) {
      const formItem = EditableElement.getFormItem();
      const ckeditor5Id = formItem.querySelector('.edit-plus-inline-edit').dataset.ckeditor5Id;
      if (ckeditor5Id) {
        const editor = Drupal.EditPlus.Ckeditor5Instances.get(ckeditor5Id);
        if (editor) {
          editor.focus();
        } else {
          // The editor hasn't been built yet because the form was just loaded.
          // Remember that this element needs focus and focus it after the editor
          // has been built.
          Drupal.EditPlus.FocusInlineEditor = ckeditor5Id;
        }
      } else {
        // This is a text area.
        EditableElement.getFormItemInputs()[0].focus();
      }

      // Remember what the original data was so that we can update the tempstore
      // if it has changed.
      this.originalData.set(EditableElement.info.elementId, this.getInputValue(formItem));
    }

    blurElement(EditableElement) {
      const formItem = EditableElement.getFormItem();
      const textArea = EditableElement.getFormItemInputs()[0];
      const pageElement = EditableElement.getPageElement();

      const originalData = this.originalData.get(EditableElement.info.elementId);
      const data = this.getInputValue(formItem);

      if (data !== originalData) {
        pageElement.innerHTML = textArea.innerText = data;
      }
      this.replaceEditableElementWithMarkup(EditableElement);
      if (data !== originalData) {
        entityForm.updateTempstore(EditableElement, data.length === 0);
      }
    }

    revealAncillary(EditableElement) {
      if (!EditableElement.handleIsWrapper()) {
        const formItemWrapper = EditableElement.getFormItemWrapper();
        const form = formItemWrapper.closest('form');
        form.insertBefore(formItemWrapper, form.firstChild);
        formItemWrapper.classList.add('edit-plus-focused');
        formItemWrapper.classList.remove('edit-plus-hidden');
      }
    }

    getFocusedSelectors() {
      let selectors = super.getFocusedSelectors();
      selectors.push('.ck-body-wrapper');
      return selectors;
    }

    /**
     * Get data.
     *
     * Gets the user input from either the CKEditor or a textarea.
     *
     * @param formItem
     *   The form item to retrieve data from.
     *
     * @returns {*}
     *   The user input.
     */
    getInputValue(formItem) {
      const ckeditor5Id = formItem.querySelector('.edit-plus-inline-edit').dataset.ckeditor5Id;
      if (ckeditor5Id) {
        const editor = Drupal.EditPlus.Ckeditor5Instances.get(ckeditor5Id);
        if (Drupal.EditPlus.FocusInlineEditor === ckeditor5Id) {
          // The editor was the first thing that was clicked and hasn't been
          // attached to the form yet. this.originalData will be set once the
          // editor has been attached. @see attachEditor
          return null;
        } else {
          return editor.getData();
        }
      } else {
        const textArea = formItem.querySelector('[data-edit-plus-input]');
        return textArea.value;
      }
    }

  }

  /**
   * Register the InlineEditor field plugin.
   */
  window.addEventListener('EditPlusFieldManager.RegisterPlugins', e => {
    e.detail.manager.registerPlugin(new InlineEditor());
  });

  /**
   * The Inline CKEditor instances.
   */
  Drupal.EditPlus.Ckeditor5Instances = new Map();

  /**
   * Focus Inline Editor
   *
   * After replacePageMarkupWithEditableElement
   * I think it's because the inline editor is attached with a promise so we can't immediately focus it in the instance that the form item was clicked, the form loaded,
   * the editor attached, then we need to remember what was clicked so we can focus it once it has been instantiated.
   */
  Drupal.EditPlus.FocusInlineEditor = null;

  /**
   * Create Inline Editors.
   */
  Drupal.behaviors.EditPlusInlineEditors = {
    attach: (context, settings) => {
      once('EditPlusInlineEdit', '[data-inline-editor-for]', context).forEach(formatSelector => {
        // Attach the Inline Editor.
        const textAreaId = '#' + formatSelector.dataset.inlineEditorFor;
        const markupElement = document.querySelector(textAreaId).closest('.edit-plus-form-item').querySelector('.edit-plus-inline-edit');
        Drupal.behaviors.EditPlusInlineEditors.showInputElement(markupElement, formatSelector.value);

        // Listen for Editor format changes.
        formatSelector.addEventListener('change', e => {
          // Ignore the filter module changes on page load.
          if (e.isTrusted) {
            const formItem = document.querySelector('#' + e.currentTarget.dataset.inlineEditorFor).closest('.edit-plus-form-item');
            const EditableElement = editableElement.EditableElement.createFromFormItem(formItem);
            const textarea = formItem.querySelector('textarea');
            const inlineEditElement = formItem.querySelector('.edit-plus-inline-edit');
            const ckeditor5Id = inlineEditElement.dataset.ckeditor5Id;

            // Update the appropriate things before switching formats to prevent
            // data loss.
            if (ckeditor5Id) {
              const editor = Drupal.EditPlus.Ckeditor5Instances.get(ckeditor5Id);
              if (editor) {
                textarea.value = editor.getData();
              }
            } else {
              inlineEditElement.innerHTML = textarea.value;
            }

            // Remove old editor if it exists.
            if (ckeditor5Id) {
              let destroyPromise = Drupal.behaviors.EditPlusInlineEditors.detachEditor(e.currentTarget, inlineEditElement.dataset.ckeditor5Id);
              destroyPromise.then(() => {
                Drupal.behaviors.EditPlusInlineEditors.showInputElement(markupElement, formatSelector.value);
                Drupal.EditPlus.FieldPluginManager.getPlugin(EditableElement).focus(EditableElement);
              });
            } else {
              Drupal.behaviors.EditPlusInlineEditors.showInputElement(markupElement, formatSelector.value);
              Drupal.EditPlus.FieldPluginManager.getPlugin(EditableElement).focus(EditableElement);
            }
          }
        });
      });
    },

    /**
     * Show input element.
     *
     * Either reveals the textarea or builds the Inline Editor depending on the
     * field formatter.
     *
     * @param element
     *   The form element to attach the editor to.
     *
     * @param formatId
     *   The field formatter ID.
     */
    showInputElement: (element, formatId) => {
      const format = drupalSettings.editor.formats[formatId];
      if (format) {
        Drupal.behaviors.EditPlusInlineEditors.attachEditor(element, format);
      } else {
        // No editor, reveal the textarea.
        const textarea = element.closest('.edit-plus-form-item').querySelector('textarea');
        textarea.classList.remove('edit-plus-hidden');
        element.classList.add('edit-plus-hidden');
      }
    },

    /**
     * Attach Editor.
     *
     * Attaches the CKEditor Inline Editor to some markup in a form item.
     *
     * @param element
     *   The markup in a form item.
     * @param format
     *   The field formatter ID.
     */
    attachEditor: (element, format) => {
      // Hide the textarea if it hasn't been already.
      element.classList.remove('edit-plus-hidden');
      element.closest('.edit-plus-form-item').querySelector('textarea').classList.add('edit-plus-hidden');

      // Process the field formatter config.
      const { editorInline } = CKEditor5;
      const {
        toolbar,
        plugins,
        config: pluginConfig,
        language,
      } = format.editorSettings;
      const extraPlugins = core.selectPlugins(plugins);
      const config = {
        extraPlugins,
        toolbar,
        language,
        ...core.processConfig(pluginConfig),
      };

      // Add an Inline Editor
      const id = core.setElementId(element);
      const { InlineEditor } = editorInline;

      InlineEditor
        .create(element, config)
        .then((editor) => {
          Drupal.EditPlus.Ckeditor5Instances.set(id, editor);
          const formItem = editor.sourceElement.closest('.edit-plus-form-item');


          // Does this editor need focus right away? If the CKEditor was the
          // first thing that was clicked that loaded the form, we should focus
          // the editor as soon as it's created.
          if (id === Drupal.EditPlus.FocusInlineEditor) {
            Drupal.EditPlus.FocusInlineEditor = null;
            editor.focus();

            // Remember the editors original state.
            const EditableElement = editableElement.EditableElement.createFromFormItem(formItem);
            const plugin = Drupal.EditPlus.FieldPluginManager.getPlugin(EditableElement);
            // @todo review this line as it changed from getData to getInputValue
            plugin.originalData.set(EditableElement.info.elementId, plugin.getInputValue(formItem));
          }
        })
        .catch(err => {
          console.error(err.stack);
        });
    },

    /**
     * Detach Editor.
     *
     * @param formatSelector
     *   The format selector select list whose editor needs removing.
     * @param ckEditorId
     *   The CKEditor ID.
     *
     * @returns {*}
     *   A CKEditor destroy promise.
     */
    detachEditor: (formatSelector, ckEditorId) => {
      const editor = Drupal.EditPlus.Ckeditor5Instances.get(ckEditorId);
      if (editor) {
        const data = editor.getData();
        const inlineEditElement = editor.sourceElement;
        once.remove('EditPlusInlineEdit', '[data-editor-for]', formatSelector.parentElement);
        return editor.destroy()
          .then(() => {
            inlineEditElement.removeAttribute('data-ckeditor5-id');
            inlineEditElement.innerHTML = data;
          });
      }
    },
  };
}
