import * as formatterPropertyMap from './formatter-property-map.js';
import * as editableElement from './editable-element.js';

/**
 * Parse markup ID.
 *
 * @param id
 *   The Entity Plus ID on the rendered field value.
 *
 * @returns {{elementId, langcode: *, fieldName: *, mainProperty: *, delta: *, entityTypeId: *, entityId: *, widget: *, viewMode: *, fieldType: *}}
 *   The elementInfo object.
 */
export const parseMarkupId = (id) => {
  const parts = id.split('::');
  return {
    entityTypeId: parts[0],
    entityId: parts[1],
    fieldName: parts[2],
    delta: parts[3],
    langcode: parts[4],
    viewMode: parts[5],
    fieldType: parts[6],
    // widget: parts[7],
    fieldFormatter: parts[7],
    mainProperty: parts[8],
    elementId: id,
  };
}

/**
 * Parse Form Item ID.
 *
 * @param id
 *
 * @returns {{fieldName: *, delta: *, property: *, entityTypeId: *, entityId: *}}
 */
export const parseFormItemId = (id) => {
  const parts = id.split('::');
  return {
    entityTypeId: parts[0],
    entityId: parts[1],
    fieldName: parts[2],
    delta: parts[3],
    property: parts[4],
  };
}

/**
 * Get Form Item Wrapper Id.
 *
 * @param elementInfo
 *   The elementInfo object returned from parseMarkupId.
 *
 * @returns {string}
 *   The form item ID.
 */
export const getFormItemWrapperId = (elementInfo) => {
  return [
    elementInfo.entityTypeId,
    elementInfo.entityId,
    elementInfo.fieldName,
    formatterPropertyMap.getProperty(elementInfo),
  ].join('::');
}

/**
 * Get Form Item Id.
 *
 * @param elementInfo
 *   The elementInfo object returned from parseMarkupId.
 *
 * @returns {string}
 *   The form item ID.
 */
export const getFormItemId = (elementInfo) => {
  return [
    elementInfo.entityTypeId,
    elementInfo.entityId,
    elementInfo.fieldName,
    elementInfo.delta,
    formatterPropertyMap.getProperty(elementInfo),
  ].join('::');
}

/**
 * Get the elementInfo object
 *
 * @param formItem
 *   The form item.
 *
 * @returns {{elementId, langcode: *, fieldName: *, mainProperty: *, delta: *, entityTypeId: *, entityId: *, widget: *, viewMode: *, fieldType: *}}
 *   The elementInfo object.
 */
export const getElementInfo = (formItem) => {
  const markupItem = getMarkupItem(formItem);
  return parseMarkupId(markupItem.dataset.editPlusId);
}

/**
 * Get markup item.
 *
 * @param formItem
 *   The form item.
 *
 * @returns {Element|null}
 *   The markup item.
 */

// @todo Use page element instead.
export const getMarkupItem = (formItem) => {
  if (!formItem.classList.contains('edit-plus-form-item')) {
    formItem = formItem.querySelector('.edit-plus-form-item');
  }

  if (!formItem) {
    return null;
  }

  return document.querySelector('[data-edit-plus-markup-item-id="' + formItem.dataset.editPlusFormItemId + '"]');
}

/**
 * Get Form ID.
 *
 * @param elementInfo
 *   The elementInfo object returned from parseMarkupId.
 *
 * @returns {string}
 *   The Edit Plus form ID
 */
export const getFormId = (elementInfo) => {
  return [
    elementInfo.entityTypeId,
    elementInfo.entityId,
  ].join('::');
}

export const getPageElementWrapperId = (elementInfo) => {
  return [
    elementInfo.entityTypeId,
    elementInfo.entityId,
    elementInfo.fieldName,
    elementInfo.mainProperty,
  ].join('::');
}

/**
 * Edit plus update markup
 *
 * This is an AJAX method called after updating the tempstore.
 */
jQuery.fn.EditPlusUpdateMarkup = (selector, content) => {
  if (Drupal.EditPlus.onlyUpdateElement) {
    // The user made changes to one form item then clicked another editable
    // element > the next editable element was focused, the tempstore recorded the
    // change and returned the updated page markup > let's now update only the
    // first element that had its values recorded in the tempstore. If we were to
    // update the whole page at this point the actively focused form element
    // would be removed.
    const contentElement = document.createElement('div');
    contentElement.innerHTML = content;
    const elementToUpdateId = Drupal.EditPlus.onlyUpdateElement.getPageElementWrapper().dataset.editPlusFieldValueWrapper;
    const elementToUpdateSelector = '[data-edit-plus-field-value-wrapper="' + elementToUpdateId + '"]';
    const updatedElement = contentElement.querySelector(elementToUpdateSelector);
    const element = document.querySelector(elementToUpdateSelector);
    element.replaceWith(updatedElement)
    Drupal.EditPlus.onlyUpdateElement = null;
  } else {
    const entityWrapper = document.querySelector(selector);
    entityWrapper.outerHTML = content;
  }
  Drupal.EditPlus.EnableEditMode();
};
