<?php

namespace Drupal\edit_plus;

use Drupal\Core\Cache\Cache;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\AppendCommand;
use Drupal\Core\Ajax\InsertCommand;
use Drupal\Core\Ajax\RemoveCommand;
use Drupal\Core\Ajax\ReplaceCommand;
use Drupal\Core\Entity\EntityInterface;
use Drupal\edit_plus\Ajax\UpdateMarkup;
use Drupal\Core\Form\FormStateInterface;
use Drupal\edit_plus\Event\AddEmptyField;
use Drupal\Core\Render\RendererInterface;
use Drupal\edit_plus\Event\FieldAttributes;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\EntityDisplayRepositoryInterface;
use Symfony\Component\EventDispatcher\EventDispatcherInterface;
use Drupal\field_sample_value\SampleValueEntityGeneratorInterface;

trait EditPlusFormTrait {

  public $editPlusTempStore;

  public function prepareFormForInlineEditing(&$form, FormStateInterface $form_state) {
    $entity = $this->getFormEntity($form, $form_state);
    $entity = $this->editPlusTempstoreRepository->get($entity);

    // Hide the form buttons.
    $form['actions']['#attributes']['class'][] = 'edit-plus-hidden';

    // Let the JS report that a field has been emptied.
    // @see entity-form.js updateTempstore()
    $form['empty_field'] = [
      '#type' => 'hidden',
      '#value' => NULL,
      '#attributes' => [
        'id' => 'empty-field',
      ],
    ];
    foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
      if (!empty($form[$field_name])) {
        $this->addFieldToPageButton($form, $entity, $field_name);
      }
    }

    // Attribute form items so they are correlated with the rendered markup.
    $this->attributeFormItemsEventDispatcher($form, $form_state, FieldAttributes::ALTER);
    $form['#after_build'][] = [$this, 'prepareFormItemsForInlineEditingAfterBuild'];
  }

  public function prepareFormItemsForInlineEditingAfterBuild($form, FormStateInterface $form_state) {
    $this->attributeFormItemsEventDispatcher($form, $form_state, FieldAttributes::AFTER_BUILD);
    $form_state->set('added_to_page', NULL);
    return $form;
  }

  private function attributeFormItemsEventDispatcher(&$form, FormStateInterface $form_state, string $event_name) {
    $entity = $this->getFormEntity($form, $form_state);
    $entity = $this->editPlusTempstoreRepository->get($entity);
    $view_mode = $this->getViewMode($form, $form_state, $entity);
    foreach ($entity->getFieldDefinitions() as $field_name => $field_definition) {
      if (!empty($form[$field_name])) {
        $event = new FieldAttributes($form, $view_mode, $field_name, $entity, $form_state);
        $this->dispatcher->dispatch($event, $event_name);
        $form = $event->getForm();
      }
    }
  }

  /**
   * Add empty field to page button.
   *
   * Only fields with values are rendered on the canonical page, so we'd
   * have nothing to click to inline edit the field... Let's add a button in the
   * sidebar (entity form) that let's you add the field to the page with
   * auto-generated content.
   *
   * @param $form
   *   The form
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   * @param string $form_item
   *   The field name/form item to possibly add a button for if it's empty.
   */
  public function addFieldToPageButton(&$form, EntityInterface $entity, string $form_item) {
    $field = $entity->get($form_item);
    $field_definition = $field->getFieldDefinition();
    if ($field->isEmpty() && method_exists($field_definition, 'label')) {
      // Empty field items get a button to add them to the page.
      $form['empty_fields']['#type'] = 'details';
      $form['empty_fields']['#open'] = TRUE;
      $form['empty_fields']['#title'] = t('Add field');
      $form['empty_fields'][$form_item] = [
        '#type' => 'submit',
        '#value' => t('@label', ['@label' => $field_definition->label()]),
        '#name' => "add_empty_field::$form_item",
        '#submit' => [
          [$this, 'populateEmptyField'],
        ],
        '#ajax' => [
          'callback' => [$this, 'addEmptyField'],
        ],
      ];
    }
  }

  /**
   * Add empty field submit handler.
   *
   * Generate field values before adding an empty field to the page.
   */
  public function populateEmptyField(&$form, FormStateInterface $form_state) {
    $entity = $this->editPlusTempStore->get($this->getFormEntity($form, $form_state));
    $input = $form_state->getUserInput();
    [$_, $field_name] = explode('::', $input['_triggering_element_name']);
    $field = $entity->get($field_name);

    // Generate a value for the field about to be placed on the page.
    if ($this->moduleHandler->moduleExists('field_sample_value')) {
      $this->fieldGenerator->populateWithSampleValue($field);
    } else {
      $field->generateSampleItems();
    }
    $this->dispatcher->dispatch(new AddEmptyField($field), AddEmptyField::class);
    $this->updateFormState($form, $form_state, $field, $entity);
    Cache::invalidateTags([getCacheTag($entity)]);

    // Flag that the newly added field needs an "added" effect
    $form_state->set('added_to_page', $field_name);
    $form_state->setRebuild(TRUE);
  }

  /**
   * Ajax callback.
   *
   * Now that a field had its value set in populateEmptyField, update the
   * form and page.
   */
  public function addEmptyField(&$form, FormStateInterface $form_state) {
    $response = new AjaxResponse();
    $this->messenger()->addWarning($this->t('You have unsaved changes.'));
    $this->updatePage($response, $form, $form_state);
    $this->updateForm($response, $form, $form_state);
    $this->buildBottomBar($this->getMainEntity($form, $form_state), $response);
    $this->renderMessages($response);

    return $response;
  }

  /**
   * Update the rendered page.
   *
   * @param \Drupal\Core\Ajax\AjaxResponse $response
   *   The ajax response.
   * @param $form
   *   The entity form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function updatePage(AjaxResponse $response, array &$form, FormStateInterface $form_state) {
    $entity = $this->getFormEntity($form, $form_state);
    $view_mode = $this->getViewMode($form, $form_state, $entity);
    $content = $this->entityContent($form, $form_state, $view_mode);
    $selector = sprintf('[data-edit-plus-entity-wrapper="%s::%s::%s::%s"]', $entity->getEntityTypeId(), edit_plus_entity_identifier($entity), $view_mode, $entity->bundle());
    $response->addCommand(new UpdateMarkup($selector, $content));
  }

  /**
   * Update the entity form.
   *
   * @param \Drupal\Core\Ajax\AjaxResponse $response
   *   The ajax response.
   * @param $form
   *   The entity form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function updateForm(AjaxResponse $response, &$form, FormStateInterface $form_state) {
    $entity = $this->getFormEntity($form, $form_state);
    $form_id = self::getEditPlusFormId($entity);
    $response->addCommand(new ReplaceCommand("#$form_id", $this->wrapForm($form, $entity)));
  }

  /**
   * Get view display.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to get the display from.
   * @param string $view_mode
   *   The view mode for the display.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   The entity view display for the given entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getViewDisplay(EntityInterface $entity, string $view_mode) {
    $display_id = $entity->getEntityTypeId() . '.' . $entity->bundle() . '.' . $view_mode;
    static $view_displays;
    if (!empty($view_displays[$display_id])) {
      return $view_displays[$display_id];
    }
    $view_display = $this->entityTypeManager->getStorage('entity_view_display')->load($display_id);
    if (empty($view_display)) {
      $view_display = $this->entityTypeManager->getStorage('entity_view_display')->load($entity->getEntityTypeId() . '.' . $entity->bundle() . '.' . 'default');
    }
    $view_displays[$display_id] = $view_display;
    return $view_display;
  }

  public static function getEditPlusFormId(EntityInterface $entity) {
    return "edit-plus-form_{$entity->getEntityTypeId()}-{$entity->id()}";
  }

  /**
   * Wrap form.
   *
   * Wraps a form for easy identification in JS and ajax replacement. It's best
   * to wrap it after the form has been built and processed.
   *
   * @param $form
   *   The entity form to wrap.
   * @param $entity
   *   The entity.
   *
   * @return array
   *   An entity form wrapped in a container.
   */
  public function wrapForm($form, $entity) {
    return [
        // Add an AJAX wrapper.
        '#type' => 'container',
        '#tree' => FALSE,
        '#attributes' => [
          'id' => self::getEditPlusFormId($entity),
          'class' => ['edit-plus-form'],
          'data-edit-plus-form-id' => sprintf('%s::%s', $entity->getEntityTypeId(), edit_plus_entity_identifier($entity)),
        ],
        'form' => $form,
      ];
  }

  /**
   * Build bottom bar.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity to build the bottom bar for.
   * @param \Drupal\Core\Ajax\AjaxResponse $response
   *   The ajax response.
   *
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function buildBottomBar(EntityInterface $entity, AjaxResponse $response) {
    $page_bottom['bottom_bar']['edit_plus_bottom_bar'] = $this->ui->buildBottomBar([$entity->getEntityTypeId() . '.' . $entity->id() => $entity], $entity->toUrl()->toString());
    $response->addCommand(new InsertCommand('#toolbar-plus-bottom-bar', $page_bottom));
  }

  public function renderMessages(AjaxResponse $response) {
    $status_messages = ['#type' => 'status_messages'];
    $messages = $this->renderer->renderRoot($status_messages);
    if (!empty($messages)) {
      $response->addCommand(new RemoveCommand('[data-drupal-messages]'));
      $response->addCommand(new AppendCommand('.region-highlighted', $messages));
    }
  }

  /**
   * Get view mode.
   *
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *    The form state.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity.
   *
   * @return string
   *   The view mode of the entity being edited.
   */
  public function getViewMode(array &$form, FormStateInterface $form_state, EntityInterface $entity) {
    $input = $this->getUserInput($form_state);

    // View mode comes from the initial JS that loads the form or from the previously
    // loaded form for subsequent form loads. e.g. Adding an empty field to the page.
    $view_mode = $this->request->get('viewMode');
    if (!empty($view_mode)) {
      // Sanitize view mode.
      $valid_view_modes = $this->entityDisplayRepository->getViewModes($entity->getEntityTypeId());
      $valid_view_modes['default'] = TRUE;
      if (empty($valid_view_modes[$view_mode])) {
        throw new \InvalidArgumentException((string) t('Edit +: Invalid view mode for @label', ['@label' => $entity->label()]));
      }
    } else {
      $view_mode = $input['view_mode'] ?? NULL;
    }
    if (empty($view_mode)) {
      $view_mode = 'full';
    }

    // Pass the view mode along to subsequent page requests.
    if (empty($form['view_mode']['#value'])) {
      $form['view_mode'] = [
        '#type' => 'hidden',
        '#value' => $view_mode,
      ];
    }

    return $view_mode;
  }

}
